@@ -1092,8 +1092,40 @@ extern void *nvidia_stack_t_cache;
     kmem_cache_create(name, size, align, flags, ctor, NULL)
 #endif
 
+/*
+ * On Linux, when a kmem cache is created, a new sysfs entry is created
+ * for the same unless it's merged with an existing cache. Upstream Linux
+ * kernel commit 3b7b314053d021601940c50b07f5f1423ae67e21 (version 4.12+)
+ * made cache destruction asynchronous which creates a race between cache
+ * destroy and create. A new cache created with attributes as
+ * a previous cache, which is scheduled for destruction, can try to create
+ * a sysfs entry with the same conflicting name. Upstream Linux kernel
+ * commit d50d82faa0c964e31f7a946ba8aba7c715ca7ab0 (4.18) fixes this issue
+ * by cleaning up sysfs entry within slab_mutex, so the entry is deleted
+ * before a cache with the same attributes could be created.
+ *
+ * To workaround this kernel issue we take two steps:
+ * - Create unmergeable caches: a kmem_cache with a constructor is
+ *   unmergeable. So, we define an empty contructor for the same.
+ *   We need to create unmergeable caches to control all kmem destory
+ *   calls so we can flush.
+ * - Issue flush_scheduled_work() after kmem_cache_destroy() to make cache
+ *   destruction immediate and avoid race between destroy and create as
+ *   described above.
+ */
+#if defined(NV_KMEM_CACHE_HAS_KOBJ_REMOVE_WORK) && !defined(NV_SYSFS_SLAB_UNLINK_PRESENT)
+static inline void nv_kmem_ctor_dummy(void *arg)
+{
+    (void)arg;
+}
+#define NV_KMEM_CACHE_DESTROY_FLUSH flush_scheduled_work
+#else
+#define nv_kmem_ctor_dummy NULL
+#define NV_KMEM_CACHE_DESTROY_FLUSH()
+#endif
+
 #define NV_KMEM_CACHE_CREATE(name, type)    \
-    NV_KMEM_CACHE_CREATE_FULL(name, sizeof(type), 0, 0, NULL)
+    NV_KMEM_CACHE_CREATE_FULL(name, sizeof(type), 0, 0, nv_kmem_ctor_dummy)
 
 // Added in 4.16 via commit 8eb8284b412906181357c2b0110d879d5af95e52
 #if defined(NV_KMEM_CACHE_CREATE_USERCOPY_PRESENT)
@@ -1105,13 +1137,38 @@ extern void *nvidia_stack_t_cache;
 #endif
 
 #define NV_KMEM_CACHE_DESTROY(kmem_cache)   \
-    kmem_cache_destroy(kmem_cache)
+    if (kmem_cache != NULL)                 \
+    {                                       \
+        kmem_cache_destroy(kmem_cache);     \
+        NV_KMEM_CACHE_DESTROY_FLUSH();      \
+    }
 
 #define NV_KMEM_CACHE_ALLOC(kmem_cache)     \
     kmem_cache_alloc(kmem_cache, GFP_KERNEL)
 #define NV_KMEM_CACHE_FREE(ptr, kmem_cache) \
     kmem_cache_free(kmem_cache, ptr)
 
+static inline void *nv_kmem_cache_zalloc(struct kmem_cache *k, gfp_t flags)
+{
+#if defined(NV_KMEM_CACHE_HAS_KOBJ_REMOVE_WORK) && !defined(NV_SYSFS_SLAB_UNLINK_PRESENT)
+    /*
+     * We cannot call kmem_cache_zalloc directly as it adds the __GFP_ZERO
+     * flag. This flag together with the presence of a slab constructor is
+     * flagged as a potential bug by the Linux kernel since it is the role
+     * of a constructor to fill an allocated object with the desired
+     * pattern. In our case, we specified a (dummy) constructor as a
+     * workaround for a bug and not to zero-initialize objects. So, we take
+     * the pain here to memset allocated object ourselves.
+     */
+    void *object = kmem_cache_alloc(k, flags);
+    if (object)
+        memset(object, 0, kmem_cache_size(k));
+    return object;
+#else
+    return kmem_cache_zalloc(k, flags);
+#endif
+}
+
 static inline int nv_kmem_cache_alloc_stack(nvidia_stack_t **stack)
 {
     nvidia_stack_t *sp = NULL;
